@@ -52,6 +52,70 @@ class CL(object):
         _lib: handle to cffi.FFI object.
         _handle: cffi handle to OpenCL object.
     """
+
+    ERRORS = {
+        0:   "CL_SUCCESS",
+        - 1:  "CL_DEVICE_NOT_FOUND",
+        - 2:  "CL_DEVICE_NOT_AVAILABLE",
+        - 3:  "CL_COMPILER_NOT_AVAILABLE",
+        - 4:  "CL_MEM_OBJECT_ALLOCATION_FAILURE",
+        - 5:  "CL_OUT_OF_RESOURCES",
+        - 6:  "CL_OUT_OF_HOST_MEMORY",
+        - 7:  "CL_PROFILING_INFO_NOT_AVAILABLE",
+        - 8:  "CL_MEM_COPY_OVERLAP",
+        - 9:  "CL_IMAGE_FORMAT_MISMATCH",
+        - 10: "CL_IMAGE_FORMAT_NOT_SUPPORTED",
+        - 11: "CL_BUILD_PROGRAM_FAILURE",
+        - 12: "CL_MAP_FAILURE",
+        - 13: "CL_MISALIGNED_SUB_BUFFER_OFFSET",
+        - 14: "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST",
+        - 15: "CL_COMPILE_PROGRAM_FAILURE",
+        - 16: "CL_LINKER_NOT_AVAILABLE",
+        - 17: "CL_LINK_PROGRAM_FAILURE",
+        - 18: "CL_DEVICE_PARTITION_FAILED",
+        - 19: "CL_KERNEL_ARG_INFO_NOT_AVAILABLE",
+
+        - 30: "CL_INVALID_VALUE",
+        - 31: "CL_INVALID_DEVICE_TYPE",
+        - 32: "CL_INVALID_PLATFORM",
+        - 33: "CL_INVALID_DEVICE",
+        - 34: "CL_INVALID_CONTEXT",
+        - 35: "CL_INVALID_QUEUE_PROPERTIES",
+        - 36: "CL_INVALID_COMMAND_QUEUE",
+        - 37: "CL_INVALID_HOST_PTR",
+        - 38: "CL_INVALID_MEM_OBJECT",
+        - 39: "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR",
+        - 40: "CL_INVALID_IMAGE_SIZE",
+        - 41: "CL_INVALID_SAMPLER",
+        - 42: "CL_INVALID_BINARY",
+        - 43: "CL_INVALID_BUILD_OPTIONS",
+        - 44: "CL_INVALID_PROGRAM",
+        - 45: "CL_INVALID_PROGRAM_EXECUTABLE",
+        - 46: "CL_INVALID_KERNEL_NAME",
+        - 47: "CL_INVALID_KERNEL_DEFINITION",
+        - 48: "CL_INVALID_KERNEL",
+        - 49: "CL_INVALID_ARG_INDEX",
+        - 50: "CL_INVALID_ARG_VALUE",
+        - 51: "CL_INVALID_ARG_SIZE",
+        - 52: "CL_INVALID_KERNEL_ARGS",
+        - 53: "CL_INVALID_WORK_DIMENSION",
+        - 54: "CL_INVALID_WORK_GROUP_SIZE",
+        - 55: "CL_INVALID_WORK_ITEM_SIZE",
+        - 56: "CL_INVALID_GLOBAL_OFFSET",
+        - 57: "CL_INVALID_EVENT_WAIT_LIST",
+        - 58: "CL_INVALID_EVENT",
+        - 59: "CL_INVALID_OPERATION",
+        - 60: "CL_INVALID_GL_OBJECT",
+        - 61: "CL_INVALID_BUFFER_SIZE",
+        - 62: "CL_INVALID_MIP_LEVEL",
+        - 63: "CL_INVALID_GLOBAL_WORK_SIZE",
+        - 64: "CL_INVALID_PROPERTY",
+        - 65: "CL_INVALID_IMAGE_DESCRIPTOR",
+        - 66: "CL_INVALID_COMPILER_OPTIONS",
+        - 67: "CL_INVALID_LINKER_OPTIONS",
+        - 68: "CL_INVALID_DEVICE_PARTITION_COUNT",
+    }
+
     def __init__(self):
         self._lib = cl.lib  # to hold the reference
         self._handle = None
@@ -93,6 +157,14 @@ class CL(object):
             wait_list = cl.NULL
         return (wait_list, n_events)
 
+    @staticmethod
+    def get_error_name_from_code(code):
+        return CL.ERRORS.get(code, "UNKNOWN")
+
+    @staticmethod
+    def get_error_description(code):
+        return "%s (%d)" % (CL.get_error_name_from_code(code), code)
+
 
 class Event(CL):
     """Holds OpenCL event.
@@ -187,7 +259,7 @@ class Queue(CL):
         super(Queue, self).__init__()
         self._context = context
         self._device = device
-        err = cl.ffi.new("cl_int[]", 1)
+        err = cl.ffi.new("cl_int *")
         self._handle = self._lib.clCreateCommandQueue(
             context.handle, device.handle, flags, err)
         if err[0]:
@@ -273,7 +345,7 @@ class Queue(CL):
                           ptr - pointer to the mapped buffer
                                 (cffi void* converted to int).
         """
-        err = cl.ffi.new("cl_int[]", 1)
+        err = cl.ffi.new("cl_int *")
         event = cl.ffi.new("cl_event[]", 1) if need_event else cl.NULL
         wait_list, n_events = CL.get_wait_list(wait_for)
         ptr = self._lib.clEnqueueMapBuffer(
@@ -366,14 +438,16 @@ class Queue(CL):
         """
         n = self._lib.clFlush(self.handle)
         if n:
-            raise CLRuntimeError("clFlush() failed with error %d" % (n), n)
+            raise CLRuntimeError("clFlush() failed with error %s" %
+                                 CL.get_error_description(n), n)
 
     def finish(self):
         """Waits for all previous commands issued to this queue to end.
         """
         n = self._lib.clFinish(self.handle)
         if n:
-            raise CLRuntimeError("clFinish() failed with error %d" % (n), n)
+            raise CLRuntimeError("clFinish() failed with error %s" %
+                                 CL.get_error_description(n), n)
 
     def release(self):
         if self.handle is not None:
@@ -401,7 +475,7 @@ class Buffer(CL):
         self._host_array = (host_array if flags & cl.CL_MEM_USE_HOST_PTR != 0
                             else None)
         host_ptr, size = CL.extract_ptr_and_size(host_array, size)
-        err = cl.ffi.new("cl_int[]", 1)
+        err = cl.ffi.new("cl_int *")
         self._handle = self._lib.clCreateBuffer(
             context.handle, flags, size, host_ptr, err)
         if err[0]:
@@ -454,11 +528,19 @@ class Kernel(CL):
         program: Program object associated with this kernel.
         name: kernel name in the program.
     """
+
+    CL_KERNEL_FUNCTION_NAME = 0x1190
+    CL_KERNEL_NUM_ARGS = 0x1191
+    CL_KERNEL_REFERENCE_COUNT = 0x1192
+    CL_KERNEL_CONTEXT = 0x1193
+    CL_KERNEL_PROGRAM = 0x1194
+    CL_KERNEL_ATTRIBUTES = 0x1195
+
     def __init__(self, program, name):
         super(Kernel, self).__init__()
         self._program = program
         self._name = name
-        err = cl.ffi.new("cl_int[]", 1)
+        err = cl.ffi.new("cl_int *")
         ss = cl.ffi.new("char[]", name.encode("utf-8"))
         self._handle = self._lib.clCreateKernel(program.handle, ss, err)
         if err[0]:
@@ -480,6 +562,24 @@ class Kernel(CL):
         """
         return self._name
 
+    @property
+    def reference_count(self):
+        buf = cl.ffi.new("cl_uint *")
+        self._get_kernel_info(Kernel.CL_KERNEL_REFERENCE_COUNT, buf)
+        return buf[0]
+
+    @property
+    def num_args(self):
+        buf = cl.ffi.new("size_t *")
+        self._get_kernel_info(Kernel.CL_KERNEL_NUM_ARGS, buf)
+        return buf[0]
+
+    @property
+    def attributes(self):
+        buf = cl.ffi.new("char[]", 4096)
+        self._get_kernel_info(Kernel.CL_KERNEL_ATTRIBUTES, buf)
+        return cl.ffi.string(buf).decode("utf-8", "replace").strip()
+
     def set_arg(self, idx, vle, size=None):
         """Sets kernel argument.
 
@@ -518,11 +618,27 @@ class Kernel(CL):
             raise CLRuntimeError("clSetKernelArg() failed with "
                                  "error %d" % (n), n)
 
+    def set_args(self, *args):
+        for i, arg in enumerate(args):
+            if isinstance(arg, tuple) and len(arg) == 2:
+                self.set_arg(i, *arg)
+            else:
+                self.set_arg(i, arg)
+
     def release(self):
         if self.handle is not None:
             self._lib.clReleaseKernel(self.handle)
             self._handle = None
 
+    def _get_kernel_info(self, code, buf):
+        sz = cl.ffi.new("size_t *")
+        err = self._lib.clGetKernelInfo(self.handle, code,
+                                        cl.ffi.sizeof(buf), buf, sz)
+        if err:
+            raise CLRuntimeError("clGetKernelInfo() failed with error %s\n" %
+                                 CL.get_error_description(err), err)
+        return sz[0]
+
     def __del__(self):
         self.release()
 
@@ -537,31 +653,33 @@ class Program(CL):
         src: program source.
         include_dirs: list of include dirs.
         options: additional build options.
+        binary: False if the program should be created from source; otherwise,
+                src is interpreted as precompiled binaries iterable.
     """
-    def __init__(self, context, devices, src, include_dirs=(), options=""):
+
+    CL_PROGRAM_REFERENCE_COUNT = 0x1160
+    CL_PROGRAM_CONTEXT = 0x1161
+    CL_PROGRAM_NUM_DEVICES = 0x1162
+    CL_PROGRAM_DEVICES = 0x1163
+    CL_PROGRAM_SOURCE = 0x1164
+    CL_PROGRAM_BINARY_SIZES = 0x1165
+    CL_PROGRAM_BINARIES = 0x1166
+    CL_PROGRAM_NUM_KERNELS = 0x1167
+    CL_PROGRAM_KERNEL_NAMES = 0x1168
+
+    def __init__(self, context, devices, src, include_dirs=(), options="",
+                 binary=False):
         super(Program, self).__init__()
         self._context = context
         self._devices = devices
-        self._src = src
+        self._src = src.encode("utf-8") if not binary else None
         self._include_dirs = list(include_dirs)
-        self._options = options
+        self._options = options.strip().encode("utf-8")
         self._build_logs = []
-        err = cl.ffi.new("cl_int[]", 1)
-        ss = cl.ffi.new("char[]", src.encode("utf-8"))
-        strings = cl.ffi.new("char*[]", 1)
-        strings[0] = cl.ffi.cast("char*", ss)
-        self._handle = self._lib.clCreateProgramWithSource(
-            context.handle, 1, strings, cl.NULL, err)
-        if err[0]:
-            self._handle = None
-            raise CLRuntimeError("clCreateProgramWithSource() failed with "
-                                 "error %d" % (err[0]), err[0])
-        for dirnme in include_dirs:
-            if not len(dirnme):
-                continue
-            options += " -I " + (dirnme if dirnme.find(" ") < 0
-                                 else "\'%s\'" % (dirnme))
-        self._build_program(devices, options.strip().encode("utf-8"))
+        if not binary:
+            self._create_program_from_source()
+        else:
+            self._create_program_from_binary(src)
 
     @property
     def context(self):
@@ -573,66 +691,160 @@ class Program(CL):
     @property
     def devices(self):
         """
-        list of Device objects associated with this program.
+        List of Device objects associated with this program.
         """
         return self._devices
 
     @property
     def build_logs(self):
         """
-        list of program build logs (same length as devices list).
+        List of program build logs (same length as devices list).
         """
         return self._build_logs
 
     @property
-    def src(self):
+    def source(self):
         """
-        program source.
+        Program source.
         """
         return self._src
 
     @property
     def include_dirs(self):
         """
-        list of include dirs.
+        List of include dirs.
         """
         return self._include_dirs
 
     @property
     def options(self):
         """
-        additional build options.
+        Additional build options.
         """
         return self._options
 
+    @property
+    def reference_count(self):
+        buf = cl.ffi.new("cl_uint *")
+        self._get_program_info(Program.CL_PROGRAM_REFERENCE_COUNT, buf)
+        return buf[0]
+
+    @property
+    def num_kernels(self):
+        buf = cl.ffi.new("size_t *")
+        self._get_program_info(Program.CL_PROGRAM_NUM_KERNELS, buf)
+        return buf[0]
+
+    @property
+    def kernel_names(self):
+        buf = cl.ffi.new("char[]", 4096)
+        self._get_program_info(Program.CL_PROGRAM_KERNEL_NAMES, buf)
+        names = cl.ffi.string(buf).decode("utf-8", "replace")
+        return names.split(';')
+
+    @property
+    def binaries(self):
+        buf = cl.ffi.new("size_t[]", len(self.devices))
+        self._get_program_info(Program.CL_PROGRAM_BINARY_SIZES, buf)
+        sizes = list(buf)
+        buf = cl.ffi.new("char *[]", len(self.devices))
+        for i in range(len(self.devices)):
+            buf[i] = cl.ffi.new("char[]", sizes[i])
+        self._get_program_info(Program.CL_PROGRAM_BINARIES, buf)
+        bins = []
+        for i in range(len(self.devices)):
+            bins.append(cl.ffi.buffer(buf[i], sizes[i])[:])
+        return bins
+
     def get_kernel(self, name):
         """Returns Kernel object from its name.
         """
         return Kernel(self, name)
 
-    def _build_program(self, devices, options):
-        n_devices = len(devices)
+    def _get_program_info(self, code, buf):
+        sz = cl.ffi.new("size_t *")
+        err = self._lib.clGetProgramInfo(self.handle, code,
+                                         cl.ffi.sizeof(buf), buf, sz)
+        if err:
+            raise CLRuntimeError("clGetProgramInfo() failed with error %s\n" %
+                                 CL.get_error_description(err), err)
+        return sz[0]
+
+    def _create_program_from_source(self):
+        err = cl.ffi.new("cl_int *")
+        srcptr = cl.ffi.new("char[]", self.source)
+        strings = cl.ffi.new("char*[]", 1)
+        strings[0] = srcptr
+        self._handle = self._lib.clCreateProgramWithSource(
+            self.context.handle, 1, strings, cl.NULL, err)
+        del srcptr
+        if err[0]:
+            self._handle = None
+            raise CLRuntimeError("clCreateProgramWithSource() failed with "
+                                 "error %s" %
+                                 CL.get_error_description(err[0]), err[0])
+        options = self.options.decode("utf-8")
+        for dirnme in self.include_dirs:
+            if not len(dirnme):
+                continue
+            options += " -I " + (dirnme if dirnme.find(" ") < 0
+                                 else "\'%s\'" % dirnme)
+        options = options.encode("utf-8")
+        n_devices = len(self.devices)
         device_list = cl.ffi.new("cl_device_id[]", n_devices)
-        for i, dev in enumerate(devices):
+        for i, dev in enumerate(self.devices):
             device_list[i] = dev.handle
         n = self._lib.clBuildProgram(self.handle, n_devices, device_list,
                                      options, cl.NULL, cl.NULL)
         del self.build_logs[:]
         log = cl.ffi.new("char[]", 65536)
-        sz = cl.ffi.new("size_t[]", 1)
+        sz = cl.ffi.new("size_t *")
         for dev in device_list:
             m = self._lib.clGetProgramBuildInfo(
-                self.handle, dev, cl.CL_PROGRAM_BUILD_LOG, 65536, log, sz)
+                self.handle, dev, cl.CL_PROGRAM_BUILD_LOG, cl.ffi.sizeof(log),
+                log, sz)
             if m or sz[0] <= 0:
                 self.build_logs.append("")
                 continue
-            self.build_logs.append(
-                (b"".join(log[0:sz[0] - 1])).decode("utf-8", "replace"))
+            self.build_logs.append(cl.ffi.string(log).decode("utf-8",
+                                                             "replace"))
         if n:
             raise CLRuntimeError(
-                "clBuildProgram() failed with error %d\n"
+                "clBuildProgram() failed with error %s\n"
                 "Logs are:\n%s\nSource was:\n%s\n" %
-                (n, "\n".join(self.build_logs), self.src), n)
+                (CL.get_error_description(n),
+                 "\n".join(self.build_logs), self.source), n)
+
+    def _create_program_from_binary(self, src):
+        count = len(self.devices)
+        if count != len(src):
+            raise ValueError("You have supplied %d binaries for %d devices" %
+                             (len(src), count))
+        device_list = cl.ffi.new("cl_device_id[]", count)
+        for i, dev in enumerate(self.devices):
+            device_list[i] = dev.handle
+        lengths = cl.ffi.new("size_t[]", count)
+        for i, b in enumerate(src):
+            lengths[i] = len(b)
+        binaries = cl.ffi.new("unsigned char *[]", count)
+        binaries_ref = []
+        for i, b in enumerate(src):
+            binaries_ref.append(cl.ffi.new("unsigned char[]", b))
+            binaries[i] = binaries_ref[i]
+        binary_status = cl.ffi.new("cl_int[]", count)
+        err = cl.ffi.new("cl_int *")
+        self._handle = self._lib.clCreateProgramWithBinary(
+            self.context.handle, len(self.devices), device_list, lengths,
+            binaries, binary_status, err)
+        del binaries_ref
+        if err[0]:
+            self._handle = None
+            statuses = [CL.get_error_name_from_code(s) for s in binary_status]
+            raise CLRuntimeError("clCreateProgramWithBinary() failed with "
+                                 "error %s; status %s" % (
+                                     CL.get_error_description(err[0]),
+                                     ", ".join(statuses)),
+                                 err[0])
 
     def release(self):
         if self.handle is not None:
@@ -658,7 +870,7 @@ class Context(CL):
         props[0] = cl.CL_CONTEXT_PLATFORM
         props[1] = cl.ffi.cast("cl_context_properties", platform.handle)
         props[2] = 0
-        err = cl.ffi.new("cl_int[]", 1)
+        err = cl.ffi.new("cl_int *")
         n_devices = len(devices)
         device_list = cl.ffi.new("cl_device_id[]", n_devices)
         for i, dev in enumerate(devices):
@@ -709,7 +921,8 @@ class Context(CL):
         """
         return Buffer(self, flags, host_array, size)
 
-    def create_program(self, src, include_dirs=(), options="", devices=None):
+    def create_program(self, src, include_dirs=(), options="", devices=None,
+                       binary=False):
         """Creates and builds OpenCL program from source
            for the supplied devices associated with this context.
 
@@ -723,7 +936,7 @@ class Context(CL):
             Program object.
         """
         return Program(self, self.devices if devices is None else devices,
-                       src, include_dirs, options)
+                       src, include_dirs, options, binary)
 
     def release(self):
         if self.handle is not None:
@@ -893,14 +1106,14 @@ class Platform(CL):
         n = self._lib.clGetDeviceIDs(handle, cl.CL_DEVICE_TYPE_ALL,
                                      0, cl.NULL, nn)
         if n:
-            raise CLRuntimeError("clGetDeviceIDs() failed "
-                                 "with error %d" % (n), n)
+            raise CLRuntimeError("clGetDeviceIDs() failed with error %s" %
+                                 CL.get_error_description(n), n)
         ids = cl.ffi.new("cl_device_id[]", nn[0])
         n = self._lib.clGetDeviceIDs(handle, cl.CL_DEVICE_TYPE_ALL,
                                      nn[0], ids, nn)
         if n:
-            raise CLRuntimeError("clGetDeviceIDs() failed "
-                                 "with error %d" % (n), n)
+            raise CLRuntimeError("clGetDeviceIDs() failed with error %s" %
+                                 CL.get_error_description(n), n)
         self._devices = list(Device(dev_id, self,
                                     "%s:%d" % (self.path, dev_num))
                              for dev_id, dev_num in zip(ids, range(len(ids))))
@@ -953,13 +1166,13 @@ class Platforms(CL):
         nn = cl.ffi.new("cl_uint[]", 1)
         n = self._lib.clGetPlatformIDs(0, cl.NULL, nn)
         if n:
-            raise CLRuntimeError("clGetPlatformIDs() failed "
-                                 "with error %d" % (n), n)
+            raise CLRuntimeError("clGetPlatformIDs() failed with error %s" %
+                                 CL.get_error_description(n), n)
         ids = cl.ffi.new("cl_platform_id[]", nn[0])
         n = self._lib.clGetPlatformIDs(nn[0], ids, nn)
         if n:
-            raise CLRuntimeError("clGetPlatformIDs() failed "
-                                 "with error %d" % (n), n)
+            raise CLRuntimeError("clGetPlatformIDs() failed with error %s" %
+                                 CL.get_error_description(n), n)
         self._platforms = list(Platform(p_id, str(p_num))
                                for p_id, p_num in zip(ids, range(len(ids))))
 
